open Genlex
open Utils

let aterms_keywords = ["("; ")"; ","; "["; "]"]

let aterms_program_lexer spec = make_lexer aterms_keywords spec

type specification = (string * int) list * string list
    * (string * (string * string) list) list
(********************************************************)
   *  (string * (string * (string * string)) list) list
(********************************************************)

let merge_spec a b =
  match (a,b) with
(*****************************)
    | (a1,a2,a3,a4), (b1,b2,b3,b4) ->
	(a1 @ b1, a2 @ b2, a3 @ b3, a4 @ b4)
(*****************************)

let rec check_term expr =
  match expr with parser
    | [< 'Int i; p = check_term >] -> (string_of_int i) ^ p
    | [< 'Ident "X"; p = check_term >] -> "*"^ p
    | [< 'Ident "and"; 'Kwd "("; p = check_term; 'Kwd ","; q = check_term; 'Kwd ")" >] -> p ^ " /\\ " ^ q
    | [< 'Ident n; p = check_term >] ->  n ^ p
    | [< 'Kwd "("; p = check_term_seq; 'Kwd ")" >] -> "(" ^ p ^ ")"
    | [< >] -> ""
and check_term_seq expr =
  match expr with parser
    | [< e = check_term; p = check_term_seq_next >] -> e ^ p
and check_term_seq_next expr =
  match expr with parser
    | [< 'Kwd ","; e = check_term_seq >] -> "," ^ e

    | [< >] -> ""

let rec parse_ops expr =
  match expr with parser
    | [< 'Ident "op"; 'Kwd "("; op = check_term; 'Kwd ",";
	 arity = check_term; 'Kwd")"; n = parse_ops_next >] ->
	(op,int_of_string arity) :: n
and parse_ops_next expr =
  match expr with parser
    | [< 'Kwd ","; s = parse_ops >] -> s
    | [< >] -> []

let rec parse_vars expr =
  match expr with parser
    | [< var = check_term; n = parse_vars_next >] ->
	var :: n
and parse_vars_next expr =
  match expr with parser
    | [< 'Kwd ","; s = parse_vars >] -> s
    | [< >] -> []


let rec parse_trs expr =
  match expr with parser
    | [< 'Ident "rule"; 'Kwd "("; rl = check_term; 'Kwd ",";
	 rr = check_term; 'Kwd")"; n = parse_trs_next >] ->
	(rl,rr) :: n

and parse_trs_next expr =
  match expr with parser
    | [< 'Kwd ","; s = parse_trs >] -> s
    | [< >] -> []

(************)
let rec parse_trs_cond expr =
  match expr with parser
    | [< 'Ident "rulecond"; 'Kwd "("; rl = check_term; 'Kwd ",";
	 rr = check_term; 'Kwd ","; cd = check_term ; 'Kwd")"; n = parse_trs_cond_next >] ->
         (rl,(rr,cd)) :: n
   
(************)
and parse_trs_cond_next expr =
  match expr with parser
    | [< 'Kwd ","; s = parse_trs_cond >] -> s
    | [< >] -> []
(*************)

let parse_op_list expr =
  match expr with parser
    | [< 'Ident "operatorList"; 'Kwd "("; 'Kwd "[";
	 ops = parse_ops; 'Kwd "]"; 'Kwd ")" >] ->
(****************************)
	(ops,[],[],[])
(****************************)
and parse_var_list expr =
  match expr with parser
    | [< 'Ident "varList"; 'Kwd "("; 'Kwd "[";
	 vars = parse_vars; 'Kwd "]"; 'Kwd ")" >] ->
(****************************)
	([],vars,[],[])
(****************************)

and parse_trs_list expr =
  match expr with parser
    | [< 'Ident "trsList"; 'Kwd "("; 'Kwd "["; 'Ident "system";
	 'Kwd "("; 'String sysname; 'Kwd ","; 'Kwd "[";
	 trs = parse_trs; 'Kwd "]"; 'Kwd ")"; 'Kwd "]"; 'Kwd ")" >] ->
(****************************)
	([],[],[(sysname,trs)],[])
(****************************)

(**************************************************)
and parse_trs_cond_list expr =
  match expr with parser
    | [< 'Ident "trsListCond"; 'Kwd "("; 'Kwd "["; 'Ident "system";
	 'Kwd "("; 'String sysname; 'Kwd ","; 'Kwd "[";
	 trs = parse_trs_cond; 'Kwd "]"; 'Kwd ")"; 'Kwd "]"; 'Kwd ")" >] ->
	([],[],[],[(sysname,trs)])
(*************************************************)

let rec parse_spec_list expr =
  match expr with parser
    | [< opl = parse_op_list; n = parse_spec_list_next >] ->
	merge_spec opl n
    | [< varl = parse_var_list; n = parse_spec_list_next >] ->
	merge_spec varl n
    | [< trsl = parse_trs_list; n = parse_spec_list_next >] ->
	merge_spec trsl n
(********************************************************************)
    | [< trslc = parse_trs_cond_list; n = parse_spec_list_next >] ->
	merge_spec trslc n
(********************************************************************)

and parse_spec_list_next expr =
  match expr with parser
    | [< 'Kwd ","; s = parse_spec_list >] ->
	s
(********************************************************************)
    | [< >] -> ([],[],[],[])
(********************************************************************)

let parse_spec expr : specification =
  match expr with parser
    | [< 'Ident "specification"; 'Kwd "("; s = parse_spec_list; 'Kwd ")" >] ->
	s
	  (* s is of the form ([(op0,arity);...],[var0;...], *)
	  (* [("R1",[rule0;...]);...]) *)

let parse_specification s =
  let chan = open_in s in
  let specification = Stream.of_channel chan
  and get_line_error s i_error =
    let i = ref 0
    and char = "0"
    and lines = ref 1
    and ic = open_in s in
      while not(i_error = !i) do
	ignore(input ic char 0 1);
	if (char.[0] = '\n') then
	  lines := !lines + 1
	else ();
	i := !i + 1;
      done;
      close_in ic;
      string_of_int !lines
  in
    try
      let spec = parse_spec (aterms_program_lexer specification) in
	close_in chan; spec
    with
      | Failure(e) ->
	  close_in chan;
	  failwith ("Error on line " ^
			  get_line_error s (Stream.count specification)
			^ " : \n\t" ^ e ^ " .")
      | _ -> 	  failwith ("Error on line " ^
			  get_line_error s (Stream.count specification))


(**************************************************)
let get_aterms_ops (a : specification) =
  match a with (a1,_,_,_) -> a1

let get_aterms_vars (a : specification) =
  match a with (_,a2,_,_) -> a2

let get_aterms_system (a : specification) sysname =
  match a with (_,_,a3,_) ->
    try List.assoc sysname a3
    with _ -> failwith ("No such system : " ^ sysname ^ " .")

let get_aterms_system_cond (a : specification) sysname =
  match a with (_,_,_,a4) ->
    try List.assoc sysname a4
    with _ -> failwith ("No such system : " ^ sysname ^ " .")
(**************************************************)
