open Genlex
open Utils
open Environment
open Import
open Syntax_tree

exception SyntaxError of exn
  
let keywords = ["("; ")"; ","; ";"; "="; "*"; "."; "-"; "/"]
let expr_keywords = ["eval"; "var"; "seq"; "cseq"; "useq"; "getn";
		     "subsn"; "add"; "op"; "mem"; "append"; "len";
		     "not"; "and"; "or"; "concat"; "symboldepth";
(*****************************************************)
		     "plus"; "minus"; "times"; 
(*****************************************************)
		     "ne"; "eq"; "ge"; "le" ; "gt"; "lt";"et"]
(*****************************************************)
  
let line_lexer l = make_lexer keywords (Stream.of_string l)
  
let program_lexer rules = make_lexer keywords rules
  
let rec parse_path path =
  match path with parser
    | [< 'Ident n; p = parse_path >] -> n ^ p
    | [< 'Kwd k when k = "." || k = "-"
	   || k = "/"; p = parse_path >] -> k ^ p
    | [< >] -> ""
	
let rec parse_comment expr =
  match expr with parser
    | [< 'Kwd k when k <> "*"; _ = parse_comment >] -> ()
    | [< 'Kwd "*"; _ = parse_comment_next >] -> ()
    | [< 'Ident _; _ = parse_comment >] -> ()
    | [< 'Int _; _ = parse_comment >] -> ()
    | [< >] -> ()
and parse_comment_next expr =
  match expr with parser
    | [< 'Kwd ")" >] -> ()
    | [< _; _ = parse_comment >] -> ()
	
let rec check_insts env insts =
  match insts with parser
    | [< i = check_inst env; 'Kwd ";"; r = check_next_insts env >] ->
	(match i with
	   | a,b,c -> Inst(a, r, b, c))
    | [< 'Kwd "("; 'Kwd "*"; _ = parse_comment >] -> Nop
and check_next_insts env insts =
  match insts with parser
    | [< i = check_insts env >] -> i
    | [< >] -> Nop
and check_inst env inst =
  let p_name = ref ""
  and line = ref 0 in
  let guard() =
    env.set_stream_line();
    p_name := env.get_program_name();
    line := env.get_line();
    true in
    match inst with parser
      | [< 'Ident "set" when guard(); 'Ident n; 'Kwd "="; e = check_expr env >] ->
	  Set(n,e), !p_name, !line
      | [< 'Ident "let" when guard(); p = check_lexical_def env; 'Ident "in"; 'Kwd "(";
	   i = check_insts env; 'Kwd ")" >] ->
	  LetIn(p,i), !p_name, !line
      | [< 'Ident "setn" when guard(); 'Kwd "("; 'Ident n; 'Kwd ","; i = check_expr env;
	   'Kwd ")"; 'Kwd "="; e = check_expr env >] ->
	  Setn(n,i,e), !p_name, !line
      | [< 'Ident "defun" when guard(); 'Ident f; 'Kwd "("; prms = get_ident_seq; 'Kwd ")";
	   'Kwd "="; 'Kwd "("; i = check_insts env; 'Kwd ")" >] ->
	  Defun(f,prms,i), !p_name, !line
      | [< 'Ident "for" when guard(); 'Ident n; e = check_range env; 'Ident "do";
	   'Kwd "("; i = check_insts env; 'Kwd ")" >] ->
	  For(n,e,i), !p_name, !line
      | [< e = check_if env >] ->
	  e, !p_name, !line
      | [< 'Ident "genrule" when guard(); 'Kwd "("; sb = check_expr env; 'Kwd ",";
	   se = check_expr env; 'Kwd ")" >] ->
	  Genrule(sb,se), !p_name, !line
(*************************************************************************)
      | [< 'Ident "genrulecond" when guard(); 'Kwd "("; sb = check_expr env; 'Kwd ",";
	   se = check_expr env;'Kwd ",";
	   sc = check_expr env; 'Kwd ")" >] ->
	  GenruleCond(sb,se,sc), !p_name, !line
(*************************************************************************)
      | [< 'Ident "import" when guard();  'Ident n >] ->
	  Import(n), !p_name, !line
      | [< 'Ident "load" when guard(); path = parse_path >] ->
	  Load(path), !p_name, !line
      | [< e = check_expr env >] ->
	  (match e with (_,p_name,line) -> Expr(e), p_name, line)
and check_expr env expr =
  match expr with parser
    | [< 'Int i >] ->
	env.set_stream_line();
	BaseConstructor(string_of_int i,true), env.get_program_name(), env.get_line()
    | [< 'Ident n when (List.mem n expr_keywords) >] ->
	env.set_stream_line();
	check_ident_keyword env n expr
    | [< 'Ident "$"; 'Ident d; p = get_params env >] ->
	env.set_stream_line();
	(match p with
	   | None -> Eval(d), env.get_program_name(), env.get_line()
	   | Some prms -> FunCall(d, prms), env.get_program_name(), env.get_line()
	)
    | [< 'Kwd "("; e = get_seq env; 'Kwd ")" >] ->
	env.set_stream_line();
	BaseTerm(None,e), env.get_program_name(), env.get_line()
    | [< 'Ident "<"; 'Ident n; 'Ident ">" >] ->
	env.set_stream_line();
	BaseConstructor("<" ^ n ^ ">",false), env.get_program_name(), env.get_line()
    | [< 'Ident "<<"; 'Ident n; 'Ident ">>" >] ->
	env.set_stream_line();
	BaseConstructor("<<" ^ n ^ ">>",false), env.get_program_name(), env.get_line()
    | [< 'Ident n >] ->
	env.set_stream_line();
	check_ident env n expr
and check_ident_keyword env n expr =
  match expr with parser
    | [< 'Kwd "(" >] ->
	(match n with
	   | "eval" ->
	       (match expr with parser
		  | [< 'Ident d; 'Kwd ")" >] ->
		      Eval(d), env.get_program_name(), env.get_line()
	       )
	   | "var" ->
	       (match expr with parser
		  | [< 'Ident v; 'Kwd ")" >] ->
		      BaseVar(v), env.get_program_name(), env.get_line()
	       )
	   | "seq" ->
	       (match expr with parser
		  | [< s = check_seq_var; 'Kwd ","
			 ?? "First operand for 'seq' must refere to the
  root of a term"; len = check_expr env; 'Kwd ")" >] ->
		      Seq(fst s, snd s, len), env.get_program_name(), env.get_line()
	       )
	   | "cseq" ->
	       (match expr with parser
		  | [< 'Ident n; 'Kwd ","; len = check_expr env;
		       'Kwd ")" >] ->
		      CSeq(n, len), env.get_program_name(), env.get_line()
	       )
	   | "useq" ->
	       (match expr with parser
		  | [< 'Ident n; 'Kwd ","; len = check_expr env;
		       'Kwd ")" >] ->
		      USeq(n, len), env.get_program_name(), env.get_line()
	       )
	   | "getn" ->
	       (match expr with parser
		  | [< e = check_expr env; 'Kwd ","; n = check_expr env;
		       'Kwd ")" >] ->
		      Getn(e, n), env.get_program_name(), env.get_line()
	       )
	   | "subsn" ->
	       (match expr with parser
		  | [< n = check_expr env; 'Kwd ","; i = check_expr env;
		       'Kwd ","; e = check_expr env; 'Kwd ")" >] ->
		      Subsn(n,i,e), env.get_program_name(), env.get_line()
	       )
	   | "add" ->
	       (match expr with parser
		  | [< i = check_expr env; 'Kwd ",";
		       e = check_expr env; 'Kwd ")" >] ->
		      Add([i;e]), env.get_program_name(), env.get_line()
	       )
	   | "op" ->
	       (match expr with parser
		  | [< op = check_expr env; 'Kwd ","; s = check_expr env;
		       'Kwd ")" >] ->
		      Op(op, s), env.get_program_name(), env.get_line()
	       )
	   | "mem" ->
	       (match expr with parser
		  | [< s = check_expr env; 'Kwd ","; d = check_expr env;
		       'Kwd ")" >] ->
		      Mem(s, d), env.get_program_name(), env.get_line()
	       )
	   | "append" ->
	       (match expr with parser
		  | [< s = check_expr env; 'Kwd ","; d = check_expr env;
		       'Kwd ")" >] ->
		      Cons(s,d), env.get_program_name(), env.get_line()
	       )
	   | "len" ->
	       (match expr with parser
		  | [< s = check_expr env; 'Kwd ")" >] ->
		      Len(s), env.get_program_name(), env.get_line()
	       )
	   | "symboldepth" ->
	       (match expr with parser
		  | [< s = check_expr env; 'Kwd ")" >] ->
		      Depth(s), env.get_program_name(), env.get_line()
	       )
		 
	   | "not" ->
	       (match expr with parser
		  | [< s = check_expr env; 'Kwd ")" >] -> 
		      Not(s), env.get_program_name(), env.get_line()
	       )
	   | "and" ->
	       (match expr with parser
		  | [< i = check_expr env; 'Kwd ",";
		       e = check_expr env; 'Kwd ")" >] ->
		      And([i;e]), env.get_program_name(), env.get_line()
	       )
	   | "or" ->
	       (match expr with parser
		  | [< i = check_expr env; 'Kwd ",";
		       e = check_expr env; 'Kwd ")" >] ->
		      Or([i;e]), env.get_program_name(), env.get_line()
	       )
	   | "concat" ->
	       (match expr with parser
		  | [< 'Ident n; 'Kwd ","; s = check_expr env;
		       'Kwd ")" >] ->
		      Concat(n,s), env.get_program_name(), env.get_line()
	       )

(***************************************************)
	   | "plus" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Plus(a,b), env.get_program_name(), env.get_line()
	       )
	   | "minus" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Minus(a,b), env.get_program_name(), env.get_line()
	       )
	   | "times" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Times(a,b), env.get_program_name(), env.get_line()
	       )
(**********************************************)
	   | "ne" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Ne(a,b), env.get_program_name(), env.get_line()
	       )
	   | "eq" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Eq(a,b), env.get_program_name(), env.get_line()
	       )
	   | "ge" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Ge(a,b), env.get_program_name(), env.get_line()
	       )
	   | "le" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Le(a,b), env.get_program_name(), env.get_line()
	       )
	   | "gt" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Gt(a,b), env.get_program_name(), env.get_line()
	       )
	   | "lt" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Lt(a,b), env.get_program_name(), env.get_line()
	       )
(**********************************************)
	   | "et" ->
	       (match expr with parser
		  | [< a = check_expr env; 'Kwd ",";
		       b = check_expr env; 'Kwd ")" >] ->
		      Et(a,b), env.get_program_name(), env.get_line()
	       )
	   | _ -> failwith "Invalid language keyword"
	)
    | [< >] -> check_ident env n expr
and check_ident env n expr =
  match expr with parser
    | [< s = check_attr env >] ->
	if (s = []) then
	  BaseConstructor(n, false), env.get_program_name(), env.get_line()
	else
	  BaseTerm(Some(n), s), env.get_program_name(), env.get_line()
and get_seq_next env seq =
  match seq with parser
    | [< 'Kwd ","; s = get_seq env >] ->
	s
    | [< >] -> []
and get_seq env seq =
  match seq with parser
    | [< e = check_expr env; s = get_seq_next env >] ->
	e :: s
    | [< >] -> []
and get_params env prms =
  match prms with parser
    | [< 'Kwd "("; s = get_seq env; 'Kwd ")" >] ->
	Some s
    | [< >] -> None
and get_ident_seq_next seq =
  match seq with parser
    | [< 'Kwd ","; s = get_ident_seq_elem >] ->
	s
    | [< >] -> []
and get_ident_seq seq =
  match seq with parser
    | [< 'Ident e; s = get_ident_seq_next >] ->
	e :: s
    | [< >] -> []
and get_ident_seq_elem seq =
  match seq with parser
    | [< 'Ident e; s = get_ident_seq_next >] ->
	e :: s
and check_attr env attr =
  match attr with parser
    | [< 'Kwd "("; e = get_seq env; 'Kwd ")" >] -> e
    | [< >] -> []
and check_range env range =
  match range with parser
    | [< 'Ident "in"; e = check_expr env >] -> Range(e)
    | [< 'Ident "from"; sn = check_expr env; 'Ident "to"; sm = check_expr env >] ->
	DblRange(sn,sm)
and check_lexical_def env expr =
  match expr with parser
    | [< 'Ident n; 'Kwd "="; e = check_expr env; s = check_lexical_def_next env >] ->
	(n,e) :: s
and check_lexical_def_next env expr =
  match expr with parser
    | [< 'Ident "and"; e = check_lexical_def env >] -> e
    | [< >] -> []
and check_seq_var expr =
  match expr with parser
    | [< 'Ident "var"; 'Kwd "("; 'Ident v; 'Kwd ")" >] ->  (v, true)
    | [< 'Ident n >] -> (n, false)
and check_if env expr =
  match expr with parser
    | [< 'Ident "if"; 'Kwd "("; comp = check_compare env;
	 'Kwd ")"; 'Ident "then"; 'Kwd "("; i = check_insts env; 'Kwd ")";
	 e = check_else env >] ->
	If(comp,i,e)
and check_else env expr =
  match expr with parser
    | [< 'Ident "else"; 'Kwd "("; i = check_insts env; 'Kwd ")" >] ->
	i
    | [< >] -> Nop
and check_compare_ops env p expr =
  match expr with parser
    | [< 'Ident "<>"; e = check_expr env >] ->
	Different(p, e), env.get_program_name(), env.get_line()
    | [< 'Kwd "="; e = check_expr env >] ->
	Equal(p, e), env.get_program_name(), env.get_line()
    | [< 'Ident "<"; e = check_expr env >] ->
	Inf(p, e), env.get_program_name(), env.get_line()
    | [< 'Ident ">"; e = check_expr env >] ->
	Sup(p, e), env.get_program_name(), env.get_line()
    | [< >] -> p
and check_compare env expr =
  match expr with parser
    | [< p = check_expr env; b1 = check_compare_ops env p;
	 (op,b2) = check_compare_next env >] ->
	match op with
	  | Some "&&" ->
	      (match b2 with
		 | Some b -> And([b1;b]), env.get_program_name(), env.get_line()
		 | None -> b1)
	  | Some "||" ->
	      (match b2 with
		 | Some b -> Or([b1;b]), env.get_program_name(), env.get_line()
		 | None -> b1)
	  | None -> b1
	  | Some c -> failwith ("Unknown comparator : " ^ c)
and check_compare_next env expr =
  match expr with parser
    | [< 'Ident "&&"; b = check_compare env >] -> (Some "&&", Some(b))
    | [< 'Ident "||"; b = check_compare env >] -> (Some "||", Some(b))
    | [< >] -> (None,None)
	
let rec exec_inst p_name line env inst =
  let string_of_expr e = string_of_base_expr (eval_expr env e) in
    match inst with
      | Import(n) ->
	  ignore(List.map (exec_inst p_name line env)
		   (List.map (fun x ->
				let (a,_,_) = (check_inst env (line_lexer x))
				in a) (import n)));
	  unit
      | Load(path) ->
	  merge_environments env
	    (let p = (env.get_current_dir()) ^ "/" ^ path in
	       if (List.mem_assoc p (get_parsed_programs ())) then
		 List.assoc p (get_parsed_programs ())
	       else
		 let penv = parse_program (p) in
		   set_parsed_program p penv;
		   penv
	    );
	  unit
      | Genrule(sb,se) ->
	  let (sb,se) = (string_of_expr sb, string_of_expr se) in
	    env.add_rule sb se; unit  
(*******************************************************)    
      | GenruleCond(sb,se,sc) ->
	  let (sb,se,sc) = (string_of_expr sb, string_of_expr se, string_of_expr sc) in
	    env.add_rule_cond sb se sc; unit
(*******************************************************)  
      | Set(n,v) -> env.define n (eval_expr env v); unit
      | Setn(n,i,v) ->
	  env.define n (eval_expr env
			  ((Subsn((Eval(n),p_name,line),i,v),p_name,line))); unit
      | Defun(n,p,c) -> env.defun n p c; unit
      | LetIn(lv,c) ->
	  env.push (List.map (fun x -> (fst x, eval_expr env (snd x))) lv);
	  let i = exec_insts env c in
	    env.pop (fst (List.split lv));
	    i
      | If(cc,cif,cel) ->
	  (match (eval_expr env cc) with
	     | Constructor(x) when x = val_false -> exec_insts env cel
	     | _ -> exec_insts env cif)
      | For(n,r,c) ->
	  let gen_inst x = LetIn([(n,x)],c) in
	    (match r with
	       | Range e ->
		   let ev_e = eval_expr env e in
		     (match ev_e with
			| Term(None,[]) -> unit
			| Term(None,l) ->
			    List.hd
			      (List.rev
				 (List.map
				    (fun x -> exec_inst p_name line env
				       (gen_inst (BaseExpr(x),p_name,line))) l))
			| _ -> failwith ("Invalid range : "
					 ^ string_of_base_expr ev_e)
		     )
	       | DblRange(a,b) ->
		   let (a0,an0) =
		     (let bound = eval_expr env a in
			(match bound with
			   | Constructor(ac) -> split_id_num ac
			   | _ -> failwith ("Invalid bound : "
					    ^ string_of_base_expr bound)))
		   and (b0,bn0) =
		     (let bound = eval_expr env b in
			(match bound with
			   | Constructor(bc) -> split_id_num bc
			   | _ -> failwith ("Invalid bound : "
					    ^ string_of_base_expr bound)))
		   in
		     if (a0 = b0) then
		       let l = ref [] in
			 for i = an0 to bn0 do
			   l := exec_inst p_name line env
			     (gen_inst ((BaseConstructor
					   (combine_id_num(a0,i),a0 = "")),p_name,line))
			   :: !l
			 done;
			 if (!l <> []) then
			   List.hd !l
			 else unit
		     else
		       failwith ("Invalid bounds : " ^ a0 ^ " <> " ^ b0)
	    )
      | Expr(e) -> eval_expr env e
and get_expr env expr =
  match expr with
    | BaseExpr(e),_,_ ->
	e
    | BaseTerm(a,b),_,_ ->
	let ev_args = List.map (fun x -> eval_expr env x) b in
	  (match a with
	     | Some t ->
		 let len = List.length ev_args in
		   if (len = 0) then
		     (env.add_op t 0; Constructor(t))
		   else if (len = 1) then
		     match ev_args with
		       | [Term(None,x)] ->
			   env.add_op t (List.length(x)); Term(a,x)
		       | _ -> env.add_op t 1; Term(a,ev_args)
		     else (env.add_op t len; Term(a,ev_args))
	     | None -> Term(a,ev_args))
    | BaseConstructor(c,b),_,_ ->
	(if (b = false) then env.add_op c 0); Constructor(c)
    | BaseVar(n),_,_ ->
	env.add_var n; Var(n)
    | Eval(n),_,_ ->
	env.get_def n
    | Seq(n,b,l),_,_ ->
	let len = int_of_base_expr (eval_expr env l)
	and lseq = ref [] in
	  for i=0 to (len - 1) do
	    let elem = n ^ (string_of_int i) in
	      if (b = true) then
		(env.add_var elem;
		 lseq := Var(elem) :: !lseq)
	      else (env.add_op elem 0; lseq := Constructor(elem) :: !lseq)
	  done;
	  Term(None,List.rev !lseq)
    | CSeq(n,l),_,_ ->
	let len = int_of_base_expr (eval_expr env l)
	and lseq = ref [] in
	  env.add_op n 0;
	  for i=0 to (len - 1) do
	    lseq := Constructor(n) :: !lseq
	  done;
	  Term(None,List.rev !lseq)
    | USeq(n,l),_,_ ->
	let len = int_of_base_expr (eval_expr env l)
	and lseq = ref [] in
	  for i=0 to (len - 1) do
	    let elem = n ^ (string_of_int i) in
	      env.add_op elem (-1);
	      lseq := Constructor(elem) :: !lseq
	  done;
	  Term(None,List.rev !lseq)
    | Getn(e,n),_,_ ->
	let e = eval_expr env e
	and n = int_of_base_expr (eval_expr env n) in
	  term_getn e n
    | Subsn(e,n,v),_,_ -> 
	let e = eval_expr env e
	and n = int_of_base_expr (eval_expr env n)
	and v = eval_expr env v in
	  term_setn e n v
    | Add(l),_,_ ->
	let head = eval_expr env (List.hd l) in
	  (match head with
	     | Constructor(c) -> let (id,num) = split_id_num c in
	       let sum = List.fold_left term_add (Constructor(string_of_int num))
		 (List.map (eval_expr env) (List.tl l)) in
		 (match sum with
		    | Constructor(c) -> Constructor(id ^ c)
		    | _ -> failwith ("Invalid operands for add"))
	     | _ -> failwith ("Invalid operand for add : " ^ string_of_base_expr head)
	  )
    | Op(r,p),_,_ ->
	(match r with
	   | Eval(n),p_name,line -> let r = env.get_def n in
	       (match r with
		  | Constructor(c) -> eval_expr env (BaseTerm(Some c,[p]),p_name,line)
		  | _ -> failwith ("Invalid parameters for op"))
	   | BaseConstructor(c,false),p_name,line ->
	       eval_expr env (BaseTerm(Some c,[p]),p_name,line)
	   | _ -> failwith ("Invalid parameters for op"))
    | Mem(p,e),_,_ ->
	let ev_p = eval_expr env p
	and ev_e = eval_expr env e in
	  (match ev_e with
	     | Term(_,b) ->
		 let res = ref val_false in
		   for i = 0 to List.length b - 1 do
		     if (List.nth b i = ev_p) then
		       res := string_of_int (i + 1)
		   done;
		   Constructor(!res)
	     | _ -> failwith ("Invalid parameters for mem"))
    | Cons(hd,tl),_,_ ->
	let ev_hd = eval_expr env hd
	and ev_tl = eval_expr env tl in
	  (match ev_tl with
	     | Term(a,b) -> Term(a,ev_hd :: b)
	     | _ -> failwith ("Invalid parameter for append : "
			      ^ string_of_base_expr ev_tl))
    | Len(e),_,_ ->
	let ev_e = eval_expr env e in
	  (match ev_e with
	     | Term(_,b) -> Constructor(string_of_int(List.length b))
	     | Constructor(_) -> Constructor(string_of_int 0)
	     | _ -> failwith ("Invalid parameter for len : "
			      ^ string_of_base_expr ev_e))
    | Depth(e),_,_ ->
	let ev_e = eval_expr env e in
	  (match ev_e with
	     | Term(Some a,_) -> Constructor(string_of_int (term_symboldepth a ev_e))
	     | _ -> Constructor(string_of_int 0))
    | Not(e),_,_ ->
	let ev_e = eval_expr env e in
	  (match ev_e with
	     | Constructor(c) when c = val_false -> Constructor(val_true)
	     | _ -> Constructor(val_false))
    | And(l),_,_ ->
	let ev_hd = eval_expr env (List.hd l)
	and ev_tl = List.map (eval_expr env) (List.tl l) in
	  List.fold_left term_and ev_hd ev_tl
    | Or(l),_,_ ->
	let ev_hd = eval_expr env (List.hd l)
	and ev_tl = List.map (eval_expr env) (List.tl l) in
	  List.fold_left term_or ev_hd ev_tl
    | Concat(s,e),p_name,line ->
	let ev_e = eval_expr env e in
	  (match ev_e with
	     | Term(None,b) ->
		 let p = List.map (fun x -> match x with
				     | Constructor(c) -> c
				     | _ -> failwith
					 ("Invalid parameter for concat : "
					  ^ string_of_base_expr x))
		   (List.rev (List.tl (List.rev b))) in
		   (match List.hd(List.rev b) with
		      | Constructor c ->
			  eval_expr env
			    (BaseConstructor(s ^ (String.concat "" p) ^ c,false),p_name,line)
		      | Term(None,t) ->
			  eval_expr env
			    (BaseTerm(Some(s ^ (String.concat "" p)),
				      List.map (fun x -> BaseExpr(x),p_name,line) t),p_name,line)
		      | Term(Some(a),t) ->
			  eval_expr env
			    (BaseTerm(Some(s ^ (String.concat "" p) ^ a),
				      List.map (fun x -> BaseExpr(x),p_name,line) t),p_name,line)
		      | _ -> failwith ("Invalid operand for concat : "
				       ^ string_of_base_expr (List.hd(List.rev b))
				      )
		   )
	     | _ -> failwith ("Invalid parameter for concat : "
			      ^ string_of_base_expr ev_e))
    | FunCall(s,values),p_name,line ->
	exec_inst p_name line env (env.apply_fun s values)
    | Equal(a,b),_,_ ->
	let ev_a = eval_expr env a
	and ev_b = eval_expr env b in
	  if (ev_a = ev_b) then Constructor(val_true)
	  else Constructor(val_false)
    | Sup(a,b),_,_ ->
	let ev_a = eval_expr env a
	and ev_b = eval_expr env b in
	  if (ev_a > ev_b) then Constructor(val_true)
	  else Constructor(val_false)
    | Inf(a,b),_,_ ->
	let ev_a = eval_expr env a
	and ev_b = eval_expr env b in
	  if (ev_a < ev_b) then Constructor(val_true)
	  else Constructor(val_false)
    | Different(a,b),_,_ ->
	let ev_a = eval_expr env a
	and ev_b = eval_expr env b in
	  if (ev_a <> ev_b) then Constructor(val_true)
	  else Constructor(val_false)
(**********************************************)
    | Plus(a,b),_,_ -> Constructor("+(" ^ (string_of_base_expr (eval_expr env a)) ^ "," ^ 
				      (string_of_base_expr (eval_expr env b)) ^ ")")
    (*| Plus(a,b) -> Term(Some "+",[(eval_expr env a);(eval_expr env b)]) *)
    (*| Minus(a,b) -> Term(Some "-",[(eval_expr env a);(eval_expr env b)]) *)
    | Minus(a,b),_,_ -> Constructor("-(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ 
				       (string_of_base_expr (eval_expr env b)) ^ ")")
    | Times(a,b),_,_ -> Constructor("X(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ 
				       (string_of_base_expr (eval_expr env b)) ^ ")")
    (*| Times(a,b) -> Term(Some "*",[(eval_expr env a);(eval_expr env b)])*)
(**********************************************)
    | Ne(a,b),_,_ -> Constructor("<>(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")
    | Eq(a,b),_,_ -> Constructor("=(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")
    | Ge(a,b),_,_ -> Constructor(">=(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")
    | Le(a,b),_,_ -> Constructor("<=(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")
    | Gt(a,b),_,_ -> Constructor(">(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")
    | Lt(a,b),_,_ -> Constructor("<(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")
(**********************************************)
    | Et(a,b),_,_ -> Constructor("and(" ^  (string_of_base_expr (eval_expr env a)) ^ "," ^ (string_of_base_expr (eval_expr env b)) ^ ")")


and eval_expr env expr =
  match expr with (s,p_name,n) ->
    try get_expr env expr
    with e -> (print_string (p_name ^ ", line " ^ (string_of_int n)
			     ^ " : Error evaluating expression "
			     ^ (keystring_of_expr s) ^ "\n");
	       raise e)
and exec_insts env insts =
  match insts with
    | Inst(inst, next, p_name, n) ->
	let r1 = 
	  try (exec_inst p_name n env inst)
	  with e -> (print_string (p_name ^ ", line " ^ (string_of_int n)
				   ^ " : Error executing instruction "
				   ^ (keystring_of_inst inst) ^ "\n");
		     raise e)
	and r2 = (exec_insts env next) in
	  if (next = Nop) then r1 else r2
    | Nop -> unit
and parse_program p =
  let env = get_new_environment() in
    env.set_current_dir (Filename.dirname p);
    let chan = open_in p in
    let rules = Stream.of_channel chan
    and get_lines p =
      let i = ref 1
      and c = ref 1
      and char = "0"
      and lines = ref 1
      and l_lines = ref []
      and ic = open_in p in
	while not(!c = 0) do
	  i := !i + 1;
	  c := input ic char 0 1;
	  if (char.[0] = '\n') then
	    (lines := !lines + 1;
	     l_lines := (!i,!lines) :: !l_lines)
	  else ()
	done;
	close_in ic;
	l_lines := (1,1) :: (List.rev ((!i+1,!lines+1) :: !l_lines));
	!l_lines
    in
      env.set_lines (get_lines p);
      env.set_stream rules;
      env.set_program_name p;
      try
	let tree = 
	  try check_insts env (program_lexer rules)
	  with e -> raise (SyntaxError e)
	in
	  (ignore(exec_insts env tree);
	   close_in chan;
	   env)
      with
	| SyntaxError e ->
	    print_string ((env.get_program_name()) ^ ", line "
			  ^ (string_of_int (env.get_line()))
			  ^" : Syntax error\n");
	    raise e
	| e -> if (!debug_mode) then print_debug env; raise e
