open JClass
open JBasics
open Genlex

let gen_stub_method access static final synchronized other mname rt pt =
  ConcreteMethod(
    {cm_signature = {ms_name = mname;
		     ms_parameters = pt;
		     ms_return_type = rt};
     cm_static = static;
     cm_final = final;
     cm_synchronized = synchronized;
     cm_strict = false;
     cm_access = access;
     cm_generic_signature = None;
     cm_bridge = false;
     cm_varargs = false;
     cm_synthetic = false;
     cm_other_flags = other;
     cm_exceptions = [];
     cm_attributes = {synthetic = false;
		      deprecated = false;
		      other = []};
     cm_implementation = Native}
  )

let gen_stub_field access static final fname t =
  {cf_signature  = {fs_name = fname;
		    fs_type = t};
   cf_generic_signature = None;
   cf_access = access;
   cf_static = static;
   cf_synthetic = false;
   cf_enum = false;
   cf_kind = if final then Final else NotFinal;
   cf_value = None;
   cf_transient = false;
   cf_other_flags = [];
   cf_attributes = {synthetic = false;
		    deprecated = false;
		    other = []}
  }

let gen_stub_class access cname super fmap mmap =
  `Class {c_name = cname;
	  c_version = {major = 1; minor = 0};
	  c_access = access;
	  c_final = false;
	  c_abstract = false;
	  c_super_class = super;
	  c_generic_signature = None;
	  c_fields = fmap;
	  c_interfaces = [];
	  c_consts = [||];
	  c_sourcefile = None;
	  c_deprecated = false;
	  c_enclosing_method = None;
	  c_source_debug_extention = None;
	  c_inner_classes = [];
	  c_synthetic = false;
	  c_enum = false;
	  c_other_flags = [];
	  c_other_attributes = [];
	  c_methods = mmap
	 }

let keywords = ["["; "]"; "{"; "}"; ";"; "."]

let stub_lexer s = make_lexer keywords s

let rec make_array_type n t =
  match n with
    | 0 -> failwith "Invalid length for building an array"
    | 1 -> TObject (TArray t)
    | _ -> TObject (TArray (make_array_type (n - 1) t))

let rec parse_type expr =
  match expr with parser
    | [< 'Ident "bool"; n = parse_array 0 >] ->
	if (n = 0) then	TBasic `Bool
	else make_array_type n (TBasic `Bool)
    | [< 'Ident "byte"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Byte
	else make_array_type n (TBasic `Byte)
    | [< 'Ident "char"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Char
	else make_array_type n (TBasic `Char)
    | [< 'Ident "double"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Double
	else make_array_type n (TBasic `Double)
    | [< 'Ident "float"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Float
	else make_array_type n (TBasic `Float)
    | [< 'Ident "int"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Int
	else make_array_type n (TBasic `Int)
    | [< 'Ident "long"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Long
	else make_array_type n (TBasic `Long)
    | [< 'Ident "short"; n = parse_array 0 >] ->
	if (n = 0) then TBasic `Short
	else make_array_type n (TBasic `Short)
    | [< cn = parse_class_name; n = parse_array 0 >] ->
	if (n = 0) then TObject (TClass cn)
	else make_array_type n (TObject (TClass cn))
and parse_return_type expr =
  match expr with parser
    | [< 'Ident "void" >] -> None
    | [< t = parse_type >] -> Some t
and parse_class_name expr =
  match expr with parser
    | [< 'Ident n; p = parse_class_name_next >] -> n :: p
and parse_class_name_next expr =
  match expr with parser
    | [< 'Kwd "."; p = parse_class_name >] -> p
    | [< >] -> []
and parse_array n expr =
  match expr with parser
    | [< 'Kwd "["; 'Kwd "]"; p = parse_array (n + 1)>] -> p
    | [< >] -> n

let parse_class_access expr =
  match expr with parser
    | [< 'Ident "public" >] -> `Public
    | [< >] -> `Default

let parse_field_method_access expr =
  match expr with parser
    | [< 'Ident "public" >] -> `Public
    | [< 'Ident "private" >] -> `Private
    | [< 'Ident "protected" >] -> `Protected
    | [< >] -> `Default

let rec parse_method_params expr =
  match expr with parser
    | [< 'Kwd "{"; l = parse_type_list; 'Kwd "}" >] -> (true,l)
    | [< >] -> (false,[])
and parse_type_list expr =
  match expr with parser
    | [< t = parse_type; p = parse_type_list_next >] -> t :: p
    | [< >] -> []
and parse_type_list_next expr =
  match expr with parser
    | [< 'Kwd ";"; p = parse_type_list >] -> p
    | [< >] -> []

let rec parse_field_method_flags expr =
  match expr with parser
    | [< 'Ident "static"; p = parse_method_flags_next >] -> `Static :: p
    | [< 'Ident "final"; p = parse_method_flags_next >] -> `Final :: p
    | [< 'Ident "synchronized"; p = parse_method_flags_next >] -> `Synchronized :: p
    | [< 'Ident "native"; p = parse_method_flags_next >] -> `Native :: p
    | [< >] -> []
and parse_method_flags_next expr =
  match expr with parser
    | [< p = parse_field_method_flags >] -> p
    | [< >] -> []

let parse_method_name expr =
  match expr with parser
    | [< 'Ident "<"; 'Ident n; 'Ident ">" >] -> "<" ^ n ^ ">"
    | [< 'Ident n >] -> n

let parse_field_method_signature expr =
  match expr with parser
    | [< acc = parse_field_method_access; f = parse_field_method_flags;
	 rt = parse_return_type; name = parse_method_name;
	 pt = parse_method_params; 'Kwd ";" >] ->
	let static = List.mem `Static f
	and final = List.mem `Final f
	and synchronized = List.mem `Synchronized f
	and other = ExtList.List.filter_map
	  (fun x -> if (x = `Native) then Some 256 else None) f in
	  if (fst pt) then
	    `Method(gen_stub_method acc static final synchronized other
		      name rt (snd pt))
	  else
	    match rt with
	      | Some t ->
		  `Field(gen_stub_field acc static final name t)
	      | None -> failwith "Invalid type void for field"
	
let rec parse_fields_methods fmap mmap expr =
  match expr with parser
    | [< 'Kwd "{"; r = parse_next_fields_methods fmap mmap >] -> r
and parse_next_fields_methods fmap mmap expr =
  match expr with parser
    | [< 'Kwd "}" >] -> (fmap,mmap)
    | [< m = parse_field_method_signature;
	 p = parse_next_fields_methods fmap mmap >] ->
	(match m with
	   | `Method m ->
	       (match m with
		  | ConcreteMethod cm ->
		      (fmap, MethodMap.add cm.cm_signature m (snd p))
		  | AbstractMethod am ->
		      (fmap, MethodMap.add am.am_signature m (snd p))
	       )
	   | `Field(cf) -> (FieldMap.add cf.cf_signature cf (fst p), mmap)
	)

let parse_super_class expr =
  match expr with parser
    | [< 'Ident "extends"; p = parse_class_name >] -> Some p
    | [< >] -> None

let parse_class expr =
  match expr with parser
    | [< acc = parse_class_access; 'Ident "class"; cname = parse_class_name;
	 super = parse_super_class;
	 (fmap,mmap) = parse_fields_methods FieldMap.empty MethodMap.empty >] ->
	(String.concat "." cname, (gen_stub_class acc cname super fmap mmap))

let rec parse_classes expr =
  match expr with parser
    | [< _ = Stream.empty >] -> []
    | [< c = parse_class; p = parse_classes >] -> c :: p

let stub_list = ref []
  
let parse_stubs file =
  let ic = open_in file in
  let s = Stream.of_channel ic in
  let p = stub_lexer s in
    try
      stub_list := parse_classes p;
      close_in ic
    with e ->
      close_in ic;
      let nc = Stream.count s in
      let ic = open_in file
      and i = ref 1
      and c = ref 1
      and char = "0"
      and lines = ref 1 in
	while (!i < (nc - 1) && not(!c = 0)) do
	  i := !i + 1;
	  c := input ic char 0 1;
	  if (char.[0] = '\n') then
	    lines := !lines + 1
	done;
	print_string (file ^ " : Syntax error on line " ^ (string_of_int !lines)
		      ^ "\n");
	raise e


let get_stubs_names () =
  let l = ref [] in
    for i = 1 to List.length !stub_list do
      l := fst (List.nth !stub_list (i - 1)) :: !l
    done;
    List.rev !l

let get_stub s = List.assoc s !stub_list
