open Import
open Interpreter
open JClassAbs
open Export
open Environment
open Utils
open Stubs
open Syntax_tree

type program_parameters =
  {mutable aterms_file : string;
   mutable recompile : bool;
   mutable system_name : string;
   mutable rules_file : string;
   mutable screen_width : int;
   mutable maude_file : string;
   mutable timbuk_file : string;
   mutable class_file : string;
   mutable classpath : string;
   mutable program : program;
   mutable stubfile : string}
    
let parameters = {aterms_file = ""; recompile = true; system_name = "S";
		  rules_file = "rules.rex"; screen_width = 100;
		  maude_file = ""; timbuk_file = ""; class_file = "";
		  classpath = ""; program = nilprogram;
		  stubfile = "stubclasses.jstub"}

let rules_action = fun x -> parameters.rules_file <- x
let rules_doc = "specifies the rules file (default : rules.rex)"
let aterms_action = fun x -> parameters.aterms_file <- x
let aterms_doc = "specifies the aterms result file"
let no_recompile_action() = parameters.recompile <- false
let no_recompile_doc = "If the aterms file is already generated, it's"
  ^ " not useful to parse the java class file again"
let maude_action = fun x -> parameters.maude_file <- x
let maude_doc = "specifies the maude result file"
let timbuk_action = fun x -> parameters.timbuk_file <- x
let timbuk_doc = "specifies the timbuk result file"
let screen_action = fun x -> set_screen_width x
let screen_doc = "specifies the max size of lines in result files"
let sysname_action = fun x -> parameters.system_name <- x
let sysname_doc = "specifies the name of the system (default : S)"
let classpath_action = fun x -> parameters.classpath <- x
let classpath_doc = "specifies the classpath where to search for .class files"
  ^ "\tMust be placed before -javaclass option"
let import_class_action = fun x -> parameters.class_file <- x
let import_class_doc = "specifies the .class file to import\n"
  ^ "\tMust not contain the .class extension"
let debug_action() = debug_mode := true
let debug_doc = "prints information useful for debugging"
let stubs_action x =
  parameters.stubfile <- x
let stubs_doc = "specifies the file where are defined stubs"
  
let usage = ("Usage : trex -aterms filename [-rules file]" ^
	       " [" ^
	       "-sysname expr] [-maude file] [-timbuk file]" ^
	       " [-screen-width size] [-javaclass file]" ^
	       " [-stubs file] [-no-recompile]" )

let main args =
  try
    Arg.parse_argv args
      [("-rules", Arg.String(rules_action), rules_doc);
       ("-aterms", Arg.String(aterms_action), aterms_doc);
       ("-maude", Arg.String(maude_action), maude_doc);
       ("-timbuk", Arg.String(timbuk_action), timbuk_doc);
       ("-sysname", Arg.String(sysname_action), sysname_doc);
       ("-classpath", Arg.String(classpath_action), classpath_doc);
       ("-javaclass", Arg.String(import_class_action), import_class_doc);
       ("-stubs", Arg.String(stubs_action), stubs_doc);
       ("-no-recompile", Arg.Unit(no_recompile_action), no_recompile_doc);
       ("-screen-width", Arg.Int(screen_action), screen_doc);
       ("-debug", Arg.Unit(debug_action), debug_doc)]
      (fun  _ -> ()) usage;
    (if (parameters.aterms_file = "") then
       failwith ("Error : Missing argument -aterms.\n" ^
		   "Type trex -help to get more help about the options")
     else
       ((if (parameters.recompile) then
	   (if (parameters.class_file <> "") then
	      (
		parse_stubs parameters.stubfile;
		parameters.program <- (full_parse parameters.class_file
					 parameters.classpath);
		set_program parameters.program
	      );
	    (let env = parse_program parameters.rules_file in
	       export env parameters.aterms_file parameters.system_name;
	       if (!debug_mode) then print_debug env
	    )
	   )
	);
	if (parameters.maude_file <> "") then
	  export_maude parameters.aterms_file parameters.system_name
	    parameters.maude_file;
	if (parameters.timbuk_file <> "") then
	  export_timbuk parameters.aterms_file parameters.system_name
	    parameters.timbuk_file
       )
    )
  with
    | Arg.Bad m | Arg.Help m -> print_string (m)
    | Failure m | Sys_error m -> print_string (m ^ "\n");;


let x = Sys.argv in (main x);;
