open Str
open Genlex

let rec disjonction (a : 'a list) (b : 'a list) =
  match a with
  | [] -> []
  | t::q -> if (List.mem t b) then (disjonction q b) else (t::(disjonction q b))
  
let split_id_num id =
  try
    let exp = regexp "\\(.*[a-zA-Z]+\\)\\([0-9]+$\\)" in
      if(string_match exp id 0) then
	(matched_group 1 id, int_of_string(matched_group 2 id))
      else
	match id.[0] with
	  | 'a'..'z' | 'A'..'Z' -> (id, 0)
	  | '0'..'9' -> (
	      try ("", int_of_string id)
	      with | _ -> failwith "Id not valid"
	    )
	  | _ -> failwith "Id not valid"
  with
    | _ -> failwith "Id not valid"

let combine_id_num (id,num) = id ^ string_of_int num

let remove_list_duplicates l =
  let rec remove_list_duplicates_terminal l l_res =
    match l with
      | [] -> List.rev l_res
      | hd::tl ->
	  if not(List.mem hd l_res) then
	    remove_list_duplicates_terminal tl (hd :: l_res)
	  else remove_list_duplicates_terminal tl l_res
  in remove_list_duplicates_terminal l []

let encaps s =  ("(" ^ s ^ ")")

let decaps e =
  (
    let keywords = [ "("; ")"; ","; "<"; ">" ]
    in let expr = make_lexer keywords (Stream.of_string e) in
      (let rec parse x =
	 match x with parser
	   | [< 'Kwd "("; p = parse_next; 'Kwd ")" ?? "Missing ')' after "
		  ^ p; _ = Stream.empty ?? "decaps error empty">]-> p
	   | [< p = parse_next >] -> p
       and parse_next x =
	 match x with parser
	   | [< 'Kwd "("; p = parse_next; 'Kwd ")" ?? "Missing ')' after " ^ p;
		s = parse_next >] -> encaps(p) ^ s
	   | [< 'Ident n; p = parse_next >] -> n ^ p
	   | [< 'Int i; p = parse_next >] -> (string_of_int i) ^ p
	   | [< 'Kwd k when k <> "(" && k <> ")"; p = parse_next >] ->
	       k ^ p
	   | [< >] -> ""
       in parse expr)
  )

let rec rename_constructors e =
  (
    let keywords = [ "<"; ">"; ","; "("; ")" ]
    in let expr = make_lexer keywords (Stream.of_string e) in
      (let rec parse x =
	 match x with parser
	   | [< 'Kwd "<"; 'Ident c ?? "Identifier expected after '<'";
		'Kwd ">" ?? "'>' expected after " ^ c; s = parse >] ->
	       (String.capitalize c) ^ s
	   | [< 'Ident n; p = parse >] -> n ^ p
	   | [< 'Int i; p = parse >] -> (string_of_int i) ^ p
	   | [< 'Kwd k; p = parse >] -> k ^ p
	   | [< >] -> ""
       in parse expr)
  )

let val_true = "val_true"

let val_false = "val_false"

let split_seq_content seq =
  let keywords = [ ","; "("; ")" ]
  in let expr = make_lexer keywords (Stream.of_string seq) in
      (let rec parse x =
	 match x with parser
	   | [< 'Ident i; s = parse_seq >] ->
	       let e = List.hd s in
		 if (String.length e > 0)
		 then
		   if (e.[0] = ',') then
		     i::s
		   else
		     (i ^ (List.hd s)) :: (List.tl s)
		 else [i]
	   | [< 'Kwd "("; n = parse_next; s = parse >] ->
	       let e = List.hd s in
		 if (String.length e > 0)
		 then
		   if (e.[0] = ')') then
		     ("(" ^ n ^ (List.hd s)) :: (List.tl s)
		   else
		     ("(" ^ n) :: s
		 else [("(" ^ n)]
	   | [< 'Int i; p = parse >] -> (string_of_int i) :: p
	   | [< 'Kwd k; p = parse >] -> k :: p
	   | [< >] -> [""]
       and parse_next x =
	 match x with parser
	   | [< 'Kwd "("; p = parse_next; 'Kwd ")";
		n = parse_next >] -> "(" ^ p ^ ")" ^ n
	   | [< 'Int i; p = parse_next >] -> (string_of_int i) ^ p
	   | [< 'Ident n; p = parse_next >] -> n ^ p
	   | [< 'Kwd k when k <> ")"; p = parse_next >] -> k ^ p
	   | [< >] -> ""
       and parse_seq x =
	 match x with parser
	   | [< 'Kwd "("; n = parse_next; s = parse >] ->
	       let e = List.hd s in
		 if (String.length e > 0)
		 then
		   if (e.[0] = ')') then
		     ("(" ^ n ^ (List.hd s)) :: (List.tl s)
		   else
		     ("(" ^ n) :: s
		 else [("(" ^ n)]
	   | [< s = parse >] -> s
       in snd (List.partition
		 (fun x -> x = "," || x = "") (parse expr)))

let maude_replace_underscores s =
  global_replace (regexp "_") "\\/" s

let timbuk_epure s =
  global_replace (regexp "_\\|<\\|>") "" s

let resolve_name_conflict e lc =
  let rec resolve_name_conflict_aux i e lc =
    if not(List.mem e lc) then e
    else
      let me = String.capitalize e in
	if not(List.mem me lc) then me
	else
	  let ie = e ^ (string_of_int i) in
	    if not(List.mem ie lc) then ie
	    else let mie = String.capitalize ie in
	      if not(List.mem mie lc) then mie
	      else
		resolve_name_conflict_aux (i+1) e lc in
    resolve_name_conflict_aux 1 e lc

let print_assoc_list title l =
  let sep = String.make 30 '=' in
    print_string ("(" ^ sep ^ title ^ sep ^ ")\n");
    List.iter (fun x -> print_string ((fst x) ^ " = " ^ (snd x) ^ "\n")) l;
    print_string ("(" ^ sep ^ sep ^ (String.make (String.length title) '=')
		  ^ ")\n")

let print_list title l =
  let sep = String.make 30 '=' in
    print_string ("(" ^ sep ^ title ^ sep ^ ")\n");
    List.iter (fun x -> print_string (x ^ "\n")) l;
    print_string ("(" ^ sep ^ sep ^ (String.make (String.length title) '=')
		  ^ ")\n")
  
let encaps_prog s =
  "<" ^ s ^ ">"

let decaps_prog s =
  let s_len = String.length s in
    if (s.[0] = '<') then
      let s_res = String.sub s 1 (s_len - 2) in
	if (s_res.[0] = '<') then
	  String.sub s_res 1 (s_len - 4)
	else s_res
    else s

